/* this handles all file IO, it is written to be as portable as I could make it
 */

#include	"includes.h"

/* error family, members and descriptions for those errors
 * generated here
 */

#define	LF		10
#define	CR		13

static char errorFamily[]="xFiles";

enum
	{
	BADLFILE
	};

static char *errorMembers[]=
	{
	"BadLFile"
	};

static char *errorDescriptions[]=
	{
	"Invalid LFILE pointer"
	};


LFILE
	*LFileHead;											/* start of linked list of LFILES */

BOOLEAN InitFiles()
/* do what is needed to init the file system
 * this will be called before any other file operations are done
 */
{
	LFileHead=NULL;										/* no open LFILES yet */
	return(TRUE);
}

void UnInitFiles()
/* undo what ever is needed at exit
 */
{
	while(LFileHead)
		{
		CloseLFile(LFileHead);							/* close all open L files */
		}
}

BOOLEAN FileExists(char *path,BOOLEAN *doesExist)
/* check to see if the file or directory at the end of path exists
 * if it does, return TRUE with doesExist set to TRUE
 * if it does not, doesExist will be set to FALSE
 * if there is a problem SetError, and return FALSE
 */
{
	struct stat
		statusBuffer;

	*doesExist=FALSE;
	if(stat(path,&statusBuffer)==0)
		{
		*doesExist=TRUE;
		return(TRUE);
		}
	else
		{
		if(errno==ENOENT)
			{
			return(TRUE);
			}
		SetStdCLibError();						/* set high level error information */
		}
	return(FALSE);
}

FILE *OpenReadFile(char *fname,int bufferSize)
/* passed a pointer to a file name, open a file for reading (in untranslated mode)
 * return a FILE, or NULL if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
FILE
	*theFile;

	if( (theFile=(FILE *)fopen(fname,"r")) )				/* open the file */
		{
		if(bufferSize)
			{
			if(setvbuf(theFile,NULL,_IOFBF,bufferSize)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		else
			{
			if(setvbuf(theFile,NULL,_IONBF,0)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		}
	else
		{
		SetStdCLibError();								/* set high level error information */
		}
	return(theFile);
}

FILE *OpenWriteFile(char *fname,int bufferSize)
/* passed a pointer to a file name, open a file for writing (untranslated mode)
 * return a FILE, or null if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
FILE
	*theFile;

	if( (theFile=(FILE *)fopen(fname,"w")) )				/* open the file */
		{
		if(bufferSize)
			{
			if(setvbuf(theFile,NULL,_IOFBF,bufferSize)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		else
			{
			if(setvbuf(theFile,NULL,_IONBF,0)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		}
	else
		{
		SetStdCLibError();								/* set high level error information */
		}
	return(theFile);
}

FILE *OpenTranslatedReadFile(char *fname,int bufferSize)
/* passed a pointer to a file name, open a file for reading (in translated mode)
 * return a FILE, or NULL if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
FILE
	*theFile;

	if( (theFile=(FILE *)fopen(fname,"r")) )				/* open the file */
		{
		if(bufferSize)
			{
			if(setvbuf(theFile,NULL,_IOFBF,bufferSize)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		else
			{
			if(setvbuf(theFile,NULL,_IONBF,0)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		}
	else
		{
		SetStdCLibError();								/* set high level error information */
		}
	return(theFile);
}

FILE *OpenTranslatedWriteFile(char *fname,int bufferSize)
/* passed a pointer to a file name, open a file for writing (translated mode)
 * return a FILE, or null if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
FILE
	*theFile;

	if( (theFile=(FILE *)fopen(fname,"w")) )				/* open the file */
		{
		if(bufferSize)
			{
			if(setvbuf(theFile,NULL,_IOFBF,bufferSize)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		else
			{
			if(setvbuf(theFile,NULL,_IONBF,0)!=0)
				{
				SetStdCLibError();						/* set high level error information */
				fclose(theFile);						/* could not set buf for the file, so close it and complain */
				theFile=(FILE *)NULL;
				}
			}
		}
	else
		{
		SetStdCLibError();								/* set high level error information */
		}
	return(theFile);
}

BOOLEAN CloseFile(FILE *fp)
/* passed a pointer to a file, close it
 * return FALSE if there is a problem
 * If FALSE is returned, GetError can be called to determine what went wrong
 */
{
	fclose(fp);												/* get rid of the file */
	return(TRUE);
}

ULONG GetFileSize(FILE *fp)
/* passed a pointer to an open file, return its size in bytes
 * This routine finds the size in a slightly screwy way.
 * The file is seeked to the end, then the position is read, finally the
 * file is seeked back to where it was on entry.
 */
{
ULONG
	current,size;

	current=ftell(fp);										/* get current position */
	fseek(fp,0L,2);											/* seek 0 bytes off end */
	size=ftell(fp);											/* position is length */
	fseek(fp,current,0);									/* put the position back */
	return(size);
}

ULONG GetFilePos(FILE *fp)
/* pass this routine a file pointer and it will return the position of the
 * file.
 */
{
	return(ULONG)(ftell(fp));
}

BOOLEAN SetFilePos(FILE *fp,ULONG seek)
/* passed a pointer and a position for a file, seek to that position
 * the seek is relative to the start of the file
 */
{
	return((BOOLEAN)(fseek(fp,seek,0)==0));
}

BOOLEAN ReadFile(FILE *fp,UBYTE *dest,ULONG count,ULONG *numRead)
/* passed a pointer to an open file, extract a string of bytes from the file
 * and return it at dest
 * fill numRead with the actual number of bytes read
 * pass back a BOOLEAN that tells if all was well during read
 * it will be FALSE for a read error
 * If FALSE is returned, GetError can be called to determine what went wrong
 * it is allowed to attempt to read past the end of file, numRead will just
 * return less than count.
 */
{
int
	numToRead,numJustRead;
BOOLEAN
	fail,done;

	(*numRead)=0;							/* no bytes read so far */
	fail=done=FALSE;
	while(!done&&count)
		{
		numToRead=Min(count,INT_MAX);		/* dont pass an int that is too big */
		if((numJustRead=fread(&(dest[(*numRead)]),1,numToRead,fp))!=numToRead)
			{
			done=TRUE;						/* no matter what, we are done now */
			if(ferror(fp))					/* if there was an error, report it */
				{
				SetStdCLibError();			/* set high level error information */
				fail=TRUE;
				}
			}
		(*numRead)+=(ULONG)numJustRead;		/* add in amount just read */
		count-=(ULONG)numJustRead;			/* this many less bytes to go */
		}
	return(!fail);
}

BOOLEAN WriteFile(FILE *fp,UBYTE *src,ULONG count,ULONG *numWritten)
/* passed a pointer to an open file, send the string of bytes given
 * out to it. If there is an error, return FALSE
 * If FALSE is returned, GetError can be called to determine what went wrong
 */
{
int
	numToWrite,numJustWritten;
BOOLEAN
	fail;

	(*numWritten)=0;							/* no bytes written so far */
	fail=FALSE;
	while(!fail&&count)
		{
		numToWrite=Min(count,INT_MAX);			/* dont pass an int that is too big */
		if((numJustWritten=fwrite(&(src[(*numWritten)]),1,numToWrite,fp))!=numToWrite)
			{
			SetStdCLibError();					/* set high level error information */
			fail=TRUE;							/* no matter what, we have failed now */
			}
		(*numWritten)+=(ULONG)numJustWritten;	/* add in amount just written */
		count-=(ULONG)numJustWritten;			/* this many less bytes to go */
		}
	return(!fail);
}

BOOLEAN GetLine(FILE *theFile,char *line,int lineLength,BOOLEAN *atEOF,BOOLEAN *overFlow)
/* read a line from the given file into the array line
 * if there is a read error, return FALSE
 * lineLength is considered to be the maximum number of characters
 * to read into the line (including the 0 terminator)
 * if the line fills before a CR is hit, overFlow will be set TRUE, and the rest of the
 * line will be read, but not stored
 * if the EOF is hit, atEOF is set TRUE
 */ 
{
int
	i;
UBYTE
	c;
BOOLEAN
	hadError;
ULONG
	numRead;
BOOLEAN
	stopReading;

	i=0;												/* index into output line */
	stopReading=hadError=(*atEOF)=(*overFlow)=FALSE;	
	while(!stopReading)
		{
		if(ReadFile(theFile,&c,(ULONG)1,&numRead))	/* get character from input */
			{
			if(numRead)								/* if something was read, check it, and add to output line */
				{
				if(c=='\n'||c=='\0')				/* found termination? */
					{
					stopReading=TRUE;				/* yes, output line is complete */
					}
				else
					{
					if(c!=LF&&c!=CR)				/* see if the character should be added to the line */
						{
						if(i<lineLength-1)			/* make sure there is room to store it */
							{
							line[i++]=c;			/* store it if there is room */
							}
						else
							{
							(*overFlow)=TRUE;		/* complain of overflow, but continue to the end */
							}
						}
					}
				}
			else
				{
				stopReading=TRUE;
				(*atEOF)=TRUE;						/* tell caller that EOF was encountered */
				}
			}
		else
			{
			stopReading=TRUE;
			hadError=TRUE;
			}
		}
	if(lineLength)
		{
		line[i]='\0';									/* terminate the line if possible */
		}
	return(!hadError);									/* return error status */
}

/* ----------------------------------------------------------------- */
/* Low level file I/O -- this is used when it is necessary to read and
 * write directly, and quickly to a binary file.
 * these routines provide the same functionality as the
 * ones above, except they do it for LFILES instead of FILES
 * ALTHOUGH IT IS ANNOYING TO HAVE MORE THAN ONE TYPE OF FILE, THIS
 * IS THE ONLY SURE WAY TO GET EFFICIENT PORTABLE FILE I/O IN C.
 * It is illegal to call things like fprintf with LFILES
 */
/* ----------------------------------------------------------------- */

static void linkLFile(LFILE *theFile)
/* link the LFile to the list
 */
{
	theFile->nextFile=LFileHead;
	LFileHead=theFile;
}

static LFILE *unLinkLFile(LFILE *theFile)
/* unlink the LFile from the list
 * return the one unlinked, or NULL if it could not be located
 */
{
LFILE
	*currentFile,*previousFile;
BOOLEAN
	found;

	previousFile=NULL;
	currentFile=LFileHead;
	found=FALSE;
	while(!found&&currentFile)
		{
		if(theFile==currentFile)
			{
			found=TRUE;
			if(previousFile)
				previousFile->nextFile=currentFile->nextFile;
			else
				LFileHead=currentFile->nextFile;
			}
		else
			{
			previousFile=currentFile;
			currentFile=currentFile->nextFile;
			}
		}
	return(currentFile);
}

LFILE *OpenLReadFile(char *fname)
/* passed a pointer to a file name, open a file for reading (in untranslated mode)
 * return an LFILE, or NULL if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
LFILE
	*theFile;

	if( (theFile=(LFILE *)MNewPtr(sizeof(LFILE))) )		/* get memory */
		{
		if((theFile->thePointer=open(fname,O_RDONLY,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))!=-1)
			{
			linkLFile(theFile);						/* it has been opened, so link it to the list of open files */
			}
		else
			{
			SetStdCLibError();						/* set high level error information */
			MDisposePtr(theFile);
			theFile=NULL;
			}
		}
	else
		{
		SetStdCLibError();							/* set high level error information */
		}
	return(theFile);
}

LFILE *OpenLUpdateFile(char *fname)
/* passed a pointer to a file name, open a file for update (in untranslated mode)
 * return an LFILE, or NULL if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
LFILE
	*theFile;

	if( (theFile=(LFILE *)MNewPtr(sizeof(LFILE))) )		/* get memory */
		{
		if((theFile->thePointer=open(fname,O_RDWR,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))!=-1)
			{
			linkLFile(theFile);						/* it has been opened, so link it to the list of open files */
			}
		else
			{
			SetStdCLibError();						/* set high level error information */
			MDisposePtr(theFile);
			theFile=NULL;
			}
		}
	else
		{
		SetStdCLibError();							/* set high level error information */
		}
	return(theFile);
}

LFILE *OpenLWriteFile(char *fname)
/* passed a pointer to a file name, open a file for writing (untranslated mode)
 * return an LFILE, or NULL if there was an error
 * If NULL is returned, GetError can be called to determine what went wrong
 */
{
LFILE
	*theFile;

	if( (theFile=(LFILE *)MNewPtr(sizeof(LFILE))) )		/* get memory */
		{
		if((theFile->thePointer=open(fname,O_CREAT|O_TRUNC|O_WRONLY,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))!=-1)
			{
			linkLFile(theFile);						/* it has been opened, so link it to the list of open files */
			}
		else
			{
			SetStdCLibError();						/* set high level error information */
			MDisposePtr(theFile);
			theFile=NULL;
			}
		}
	else
		{
		SetStdCLibError();							/* set high level error information */
		}
	return(theFile);
}

BOOLEAN CloseLFile(LFILE *fp)
/* passed a pointer to an LFILE, close it
 * return FALSE if there is a problem
 * If FALSE is returned, GetError can be called to determine what went wrong
 */
{
LFILE
	*theFile;

	if( (theFile=unLinkLFile(fp)) )
		{
		close(theFile->thePointer);					/* get rid of the file */
		MDisposePtr(theFile);
		return(TRUE);
		}
	SetError(errorFamily,errorMembers[BADLFILE],errorDescriptions[BADLFILE]);	/* if file could not be located, complain */
	return(FALSE);
}

ULONG GetLFileSize(LFILE *fp)
/* passed a pointer to an open LFILE, return its size in bytes
 */
{
ULONG
	current,size;

	current=lseek(fp->thePointer,0L,1);				/* get current position */
	size=(ULONG)lseek(fp->thePointer,0L,2);			/* seek 0 bytes off end */
	lseek(fp->thePointer,current,0);				/* put the position back */
	return(size);
}

ULONG GetLFilePos(LFILE *fp)
/* pass this routine an LFILE pointer and it will return the position of the
 * file.
 */
{
	return(ULONG)(lseek(fp->thePointer,0L,1));
}

BOOLEAN SetLFilePos(LFILE *fp,ULONG seek)
/* passed a pointer and a position for an LFILE, seek to that position
 * the seek is relative to the start of the file
 * If FALSE is returned, GetError can be called to determine what went wrong
 */
{
	if(lseek(fp->thePointer,seek,0)==-1)
		{
		SetStdCLibError();							/* set high level error information */
		return(FALSE);
		}
	return(TRUE);
}

BOOLEAN ReadLFile(LFILE *fp,UBYTE *dest,ULONG count,ULONG *numRead)
/* passed a pointer to an open file, extract a string of bytes from the file
 * and return it at dest
 * fill numRead with the actual number of bytes read
 * pass back a BOOLEAN that tells if all was well during read
 * it will be FALSE for a read error
 * If FALSE is returned, GetError can be called to determine what went wrong
 * it is allowed to attempt to read past the end of file, numRead will just
 * return less than count.
 */
{
int
	numToRead,numJustRead;
BOOLEAN
	fail,done;

	(*numRead)=0;							/* no bytes read so far */
	fail=done=FALSE;
	while(!done&&count)
		{
		numToRead=Min(count,INT_MAX);		/* dont pass an int that is too big */
		if((numJustRead=read(fp->thePointer,&(dest[(*numRead)]),numToRead))!=numToRead)
			{
			done=TRUE;						/* no matter what, we are done now */
			if(numJustRead==-1)				/* if there was an error, report it */
				{
				SetStdCLibError();			/* set high level error information */
				numJustRead=0;
				fail=TRUE;
				}
			}
		(*numRead)+=(ULONG)numJustRead;		/* add in amount just read */
		count-=(ULONG)numJustRead;			/* this many less bytes to go */
		}
	return(!fail);
}

BOOLEAN WriteLFile(LFILE *fp,UBYTE *src,ULONG count,ULONG *numWritten)
/* passed a pointer to an open file, send the string of bytes given
 * out to it. If there is an error, return FALSE
 * If FALSE is returned, GetError can be called to determine what went wrong
 */
{
int
	numToWrite,numJustWritten;
BOOLEAN
	fail;

	(*numWritten)=0;							/* no bytes written so far */
	fail=FALSE;
	while(!fail&&count)
		{
		numToWrite=Min(count,INT_MAX);			/* dont pass an int that is too big */
		if((numJustWritten=write(fp->thePointer,&(src[(*numWritten)]),numToWrite))!=numToWrite)
			{
			SetStdCLibError();					/* set high level error information */
			fail=TRUE;							/* no matter what, we have failed now */
			}
		(*numWritten)+=(ULONG)numJustWritten;	/* add in amount just written */
		count-=(ULONG)numJustWritten;			/* this many less bytes to go */
		}
	return(!fail);
}
